(*----------------------------------------------------------------------------*
 *  Direct3D sample from DirectX 9.0 SDK Summer 2003                          *
 *  Delphi adaptation by Alexey Barkovoy (e-mail: clootie@ixbt.com)           *
 *                                                                            *
 *  Delphi versions 5-7 are supported                                         *
 *                                                                            *
 *  Modified: 10-Jan-2004                                                     *
 *                                                                            *
 *  Latest version can be downloaded from:                                    *
 *     http://clootie.narod.ru/delphi                                         *
 *----------------------------------------------------------------------------*)
//-----------------------------------------------------------------------------
// File: BumpEarth.cpp
//
// Desc: Direct3D environment mapping / bump mapping sample. The technique
//       used perturbs the environment map to simulate bump mapping.
//
// Copyright (c) Microsoft Corporation. All rights reserved.
//-----------------------------------------------------------------------------

{$I DirectX.inc}

unit BumpEarthUnit;

interface

uses
  Windows, Messages, SysUtils, Math, CommDlg, Resource,
  Direct3D9, D3DX9, D3DApp, D3DFile, D3DFont, D3DUtil, DXUtil;

//{$DEFINE DEBUG_VS}  // Uncomment this line to debug vertex shaders
//{$DEFINE DEBUG_PS}  // Uncomment this line to debug pixel shaders

//-----------------------------------------------------------------------------
// Defines, constants, and global variables
//-----------------------------------------------------------------------------

// Vertex with 2nd set of tex coords (for bumpmapped environment map)
type
  PBumpVertex = ^TBumpVertex;
  TBumpVertex = packed record
    p: TD3DXVector3;
    n: TD3DXVector3;
    tu1, tv1: Single;
    tu2, tv2: Single;
  end;

  PBumpVertexArray = ^TBumpVertexArray;
  TBumpVertexArray = array [0..10000] of TBumpVertex;

const
  TBumpVertex_FVF = (D3DFVF_XYZ or D3DFVF_NORMAL or D3DFVF_TEX2);

// Converts a FLOAT to a DWORD for use in SetRenderState() calls
function F2DW(f: Single): DWORD; { return *((DWORD*)&f); }

type
  //-----------------------------------------------------------------------------
  // Name: class CMyD3DApplication
  // Desc: Application class. The base class (CD3DApplication) provides the
  //       generic functionality needed in all Direct3D samples. CMyD3DApplication
  //       adds functionality specific to this sample program.
  //-----------------------------------------------------------------------------
  CMyD3DApplication = class(CD3DApplication)
  private
    m_pFont:            CD3DFont;           // A font to output text

    m_Camera:           CModelViewerCamera; // Camera used for mouse input

    m_pBlockTexture:    IDirect3DTexture9;  // A blank, gray texture
    m_pEarthTexture:    IDirect3DTexture9;  // The Earth texture
    m_pEnvMapTexture:   IDirect3DTexture9;  // The environment map
    m_pEarthBumpTexture:IDirect3DTexture9;  // Source for the bumpmap
    m_psBumpMap:        IDirect3DTexture9;  // The actual bumpmap
    m_pSkyBox: CD3DMesh;
    m_pDispMapTexture: IDirect3DTexture9;   // Displacement map texture
    m_pDispMapBumpShader: IDirect3DVertexShader9;// Displacement mapping + bump map shader
    m_pDispMapShader: IDirect3DVertexShader9;    // Displacement mapping + no bump map shader
    m_pDispMapVertexDecl: IDirect3DVertexDeclaration9;


    m_matWorld: TD3DXMatrixA16;
    m_matView: TD3DXMatrixA16;
    m_matProj: TD3DXMatrixA16;

    m_BumpMapFormat:    TD3DFormat;         // Bumpmap texture format
    m_pEarthVB:         IDirect3DVertexBuffer9; // Geometry for the Earth
    m_dwNumSphereVertices: DWORD;

    m_bHighTesselation: Boolean;      // User options
    m_bTextureOn:       Boolean;
    m_bBumpMapOn:       Boolean;
    m_bEnvMapOn:        Boolean;
    m_bDeviceValidationFailed: Boolean;
    m_bDispMapOn:       Boolean;            // Displacement mapping on
    m_bCanDoDispMap:    Boolean;

    // Internal functions
    procedure SetMenuStates;
    function CreateEarthVertexBuffer: Hresult;
    procedure ApplyEnvironmentMap;
    function InitBumpMap: Hresult;

  protected
    function OneTimeSceneInit: HResult; override;
    function InitDeviceObjects: HResult; override;
    function RestoreDeviceObjects: HResult; override;
    function InvalidateDeviceObjects: HResult; override;
    function DeleteDeviceObjects: HResult; override;
    function Render: HResult; overload; override;
    function FrameMove: HResult; override;
    function FinalCleanup: HResult; override;
  public
    constructor Create;
    destructor Destroy; override;

    function ConfirmDevice(const pCaps: TD3DCaps9; dwBehavior: DWORD;
      adapterFormat: TD3DFormat; backBufferFormat: TD3DFormat): HRESULT; override;
    function MsgProc(hWnd: HWND; uMsg: UINT; wParam: WPARAM; lParam: LPARAM ): LRESULT; override;
 end;

implementation

function F2DW(f: Single): DWORD;
begin
  Result:= PDWord(@f)^;
end;

{ CMyD3DApplication }

//-----------------------------------------------------------------------------
// Name: CMyD3DApplication()
// Desc: Application constructor. Sets attributes for the app.
//-----------------------------------------------------------------------------
constructor CMyD3DApplication.Create;
begin
  inherited Create;

  m_strWindowTitle    := 'BumpEarth: Direct3D BumpMapping Demo';
  m_d3dEnumeration.AppUsesDepthBuffer := True;
  m_bShowCursorWhenFullscreen := True;

  m_psBumpMap         := nil;
  m_bTextureOn        := True;
  m_bBumpMapOn        := True;
  m_bEnvMapOn         := True;
  m_bHighTesselation  := True;
  m_bDispMapOn        := False;
  m_bCanDoDispMap     := False;

  m_pDispMapShader    := nil;
  m_pDispMapBumpShader:= nil;
  m_pDispMapVertexDecl:= nil;
  m_pDispMapTexture   := nil;
  m_pBlockTexture     := nil;
  m_pEarthTexture     := nil;
  m_pEarthBumpTexture := nil;
  m_pEnvMapTexture    := nil;
  m_pSkyBox           := nil;
  m_bDeviceValidationFailed := False;

  m_pFont             := CD3DFont.Create('Arial', 12, D3DFONT_BOLD);
  m_pEarthVB          := nil;

  m_Camera            := CModelViewerCamera.Create;
end;

destructor CMyD3DApplication.Destroy;
begin
  m_Camera.Free;
  inherited;
end;

//-----------------------------------------------------------------------------
// Name: OneTimeSceneInit()
// Desc: Called during initial app startup, this function performs all the
//       permanent initialization.
//-----------------------------------------------------------------------------
function CMyD3DApplication.OneTimeSceneInit: HResult;
begin
  // Set cursor to indicate that user can move the object with the mouse
{$IFDEF WIN64}
  SetClassLongPtr(m_hWnd, GCLP_HCURSOR, LoadCursor(0, IDC_SIZEALL));
{$ELSE}
  SetClassLong(m_hWnd, GCL_HCURSOR, LoadCursor(0, IDC_SIZEALL));
{$ENDIF}

  try
    m_pSkyBox := CD3DMesh.Create;
  except
    Result:= E_OUTOFMEMORY;
    Exit;
  end;

  Result:= S_OK;
end;

//-----------------------------------------------------------------------------
// Name: ApplyEnvironmentMap()
// Desc: Performs a calculation on each of the vertices' normals to determine
//       what the texture coordinates should be for the environment map.
//-----------------------------------------------------------------------------
procedure CMyD3DApplication.ApplyEnvironmentMap;
var
  matWorldView: TD3DXMatrix;
  vtx: PBumpVertex;

  dwNumSphereRings, dwNumSphereSegments: DWord;
  fDeltaRingAngle, fDeltaSegAngle: Single;

  vT: TD3DXVector4;
  fScale: Single;

  ring: Integer;
  r0, r1, y0, y1: Single;

  seg: Integer;
  x0, z0, x1, z1: Single;
begin
  // Get the World-View(WV) matrix set
  D3DXMatrixMultiply(matWorldView, m_matWorld, m_matView);

  // Lock the vertex buffer
  m_pEarthVB.Lock(0, 0, Pointer(vtx), 0 );

  // Establish constants used in sphere generation
  if m_bHighTesselation
    then dwNumSphereRings := 15
    else dwNumSphereRings := 5;
  if m_bHighTesselation
    then dwNumSphereSegments := 30
    else dwNumSphereSegments := 10;

  fDeltaRingAngle := (D3DX_PI / dwNumSphereRings);
  fDeltaSegAngle  := (2.0 * D3DX_PI / dwNumSphereSegments);

  // Generate the group of rings for the sphere
  for ring:= 0 to (dwNumSphereRings - 1) do
  begin
    r0 := Sin((ring+0) * fDeltaRingAngle);
    r1 := Sin((ring+1) * fDeltaRingAngle);
    y0 := Cos((ring+0) * fDeltaRingAngle);
    y1 := Cos((ring+1) * fDeltaRingAngle);

    // Generate the group of segments for the current ring
    for seg:= 0 to dwNumSphereSegments do 
    begin
      x0 :=  r0 * Sin(seg * fDeltaSegAngle);
      z0 :=  r0 * Cos(seg * fDeltaSegAngle);
      x1 :=  r1 * Sin(seg * fDeltaSegAngle);
      z1 :=  r1 * Cos(seg * fDeltaSegAngle);

      // Add two vertices to the strip which makes up the sphere
      // (using the transformed normal to generate texture coords)
      vtx.n   := D3DXVector3(x0, y0, z0); vtx.p := vtx.n;
      vtx.tu2 := 1.0 - seg/dwNumSphereSegments;
      vtx.tv2 := (ring+0)/dwNumSphereRings;
      D3DXVec3Transform(vT, vtx.n, matWorldView);
      fScale := 1.37 / D3DXVec4Length(vT);
      vtx.tu1 := 0.5 + fScale*vT.x;
      vtx.tv1 := 0.5 - fScale*vT.y;
      Inc(vtx);

      vtx.n   := D3DXVector3(x1,y1,z1); vtx.p := vtx.n;
      vtx.tu2 := 1.0 - seg/dwNumSphereSegments;
      vtx.tv2 := (ring+1)/dwNumSphereRings;
      D3DXVec3Transform(vT, vtx.n, matWorldView);
      fScale := 1.37 / D3DXVec4Length(vT);
      vtx.tu1 := 0.5 + fScale*vT.x;
      vtx.tv1 := 0.5 - fScale*vT.y;
      Inc(vtx);
    end;
  end;

  m_pEarthVB.Unlock;
end;

//-----------------------------------------------------------------------------
// Name: FrameMove()
// Desc: Animates the scene
//-----------------------------------------------------------------------------
function CMyD3DApplication.FrameMove: HResult;
{$WRITEABLECONST ON}
const
  fRotationAngle: Single = 0.0;
{$WRITEABLECONST OFF}
var
  mCamWorld, mRoll: TD3DXMatrix;
begin
  m_Camera.FrameMove(m_fElapsedTime);

  // Update the Earth's rotation angle
  if not m_Camera.IsBeingDragged then
    fRotationAngle:= fRotationAngle + m_fElapsedTime;

  // Setup viewing postion from ArcBall
  D3DXMatrixRotationY(mRoll, -fRotationAngle);

  mCamWorld := m_Camera.GetWorldMatrix^;

  D3DXMatrixMultiply(m_matWorld, mRoll, mCamWorld);
  m_matView := m_Camera.GetViewMatrix^;

  // Apply the environment map
  ApplyEnvironmentMap;

  Result:= S_OK;
end;

//-----------------------------------------------------------------------------
// Name: Render()
// Desc: Renders the scene.
//-----------------------------------------------------------------------------
function CMyD3DApplication.Render: HResult;
const
  c: array[0..3] of Single = (0.15,0,0,0);
var
  matWorldSkybox, matViewSkybox: TD3DXMatrixA16;
  dwNumPasses: DWORD;
  m: TD3DXMatrix;
begin
  if FAILED(m_pd3dDevice.BeginScene) then
  begin
    Result:= S_OK; // Don't Result:= a fatal error
    Exit;
  end;

  // Render the Skybox (which eliminates the need for a Clear of the device)
  begin
    D3DXMatrixScaling(matWorldSkybox, 10.0, 10.0, 10.0);
    m_pd3dDevice.SetTransform(D3DTS_WORLD, matWorldSkybox);

    matViewSkybox := m_matView;
    matViewSkybox._41 := 0.0; matViewSkybox._42 := 0.0; matViewSkybox._43 := 0.0;
    m_pd3dDevice.SetTransform(D3DTS_VIEW, matViewSkybox);

    m_pd3dDevice.SetRenderState(D3DRS_WRAP0, 0);
    m_pd3dDevice.SetSamplerState(0, D3DSAMP_MINFILTER, D3DTEXF_LINEAR);
    m_pd3dDevice.SetSamplerState(0, D3DSAMP_MAGFILTER, D3DTEXF_LINEAR);
    if ((m_d3dCaps.TextureAddressCaps and D3DPTADDRESSCAPS_MIRROR) = D3DPTADDRESSCAPS_MIRROR) then
    begin
      // Try to set it to mirror the texture address if the card can support it
      m_pd3dDevice.SetSamplerState(0, D3DSAMP_ADDRESSU, D3DTADDRESS_MIRROR);
      m_pd3dDevice.SetSamplerState(0, D3DSAMP_ADDRESSV, D3DTADDRESS_MIRROR);
    end;
    m_pd3dDevice.SetTextureStageState(0, D3DTSS_COLORARG1, D3DTA_TEXTURE);
    m_pd3dDevice.SetTextureStageState(0, D3DTSS_COLOROP,   D3DTOP_SELECTARG1);
    m_pd3dDevice.SetTextureStageState(0, D3DTSS_ALPHAARG1, D3DTA_TEXTURE);
    m_pd3dDevice.SetTextureStageState(0, D3DTSS_ALPHAOP,   D3DTOP_SELECTARG1);
    m_pd3dDevice.SetTextureStageState(1, D3DTSS_COLOROP, D3DTOP_DISABLE);
    m_pd3dDevice.SetTextureStageState(1, D3DTSS_ALPHAOP, D3DTOP_DISABLE);

    // Always pass Z-test, so we can avoid clearing color and depth buffers
    m_pd3dDevice.SetRenderState(D3DRS_ZFUNC, D3DCMP_ALWAYS);
    m_pSkyBox.Render(m_pd3dDevice);

    m_pd3dDevice.SetRenderState(D3DRS_ZFUNC, D3DCMP_LESSEQUAL);
    m_pd3dDevice.SetTransform(D3DTS_WORLD, m_matWorld);
    m_pd3dDevice.SetTransform(D3DTS_VIEW, m_matView);
  end;

  // Render earth

  m_pd3dDevice.SetRenderState(D3DRS_WRAP0, D3DWRAP_U or D3DWRAP_V);
  m_pd3dDevice.SetTextureStageState(0, D3DTSS_COLOROP, D3DTOP_MODULATE);
  m_pd3dDevice.SetTextureStageState(0, D3DTSS_COLORARG1, D3DTA_TEXTURE);
  m_pd3dDevice.SetTextureStageState(0, D3DTSS_COLORARG2, D3DTA_DIFFUSE);
  m_pd3dDevice.SetTextureStageState(0, D3DTSS_ALPHAOP, D3DTOP_SELECTARG1);
  m_pd3dDevice.SetTextureStageState(0, D3DTSS_ALPHAARG1, D3DTA_TEXTURE);
  m_pd3dDevice.SetTextureStageState(1, D3DTSS_COLOROP, D3DTOP_DISABLE);

  if m_bTextureOn
    then m_pd3dDevice.SetTexture(0, m_pEarthTexture)
    else m_pd3dDevice.SetTexture(0, m_pBlockTexture);

  if m_bDispMapOn then
  begin
    // Vertex shader is used, so texture coordinate indices should mapped
    // one to one with stage index
    m_pd3dDevice.SetTextureStageState(0, D3DTSS_TEXCOORDINDEX, 0);
    m_pd3dDevice.SetTextureStageState(1, D3DTSS_TEXCOORDINDEX, 1);
    m_pd3dDevice.SetTextureStageState(2, D3DTSS_TEXCOORDINDEX, 2);
    m_pd3dDevice.SetSamplerState(0, D3DSAMP_MIPFILTER, D3DTEXF_NONE);

    // Set displacement mapping sampler state
    m_pd3dDevice.SetSamplerState(D3DDMAPSAMPLER, D3DSAMP_MIPFILTER, D3DTEXF_NONE);
    m_pd3dDevice.SetSamplerState(D3DDMAPSAMPLER, D3DSAMP_MAGFILTER, D3DTEXF_LINEAR);
    m_pd3dDevice.SetSamplerState(D3DDMAPSAMPLER, D3DSAMP_MINFILTER, D3DTEXF_LINEAR);
    m_pd3dDevice.SetTexture(D3DDMAPSAMPLER, m_pDispMapTexture);

    // Set transformation matrix as shader constants
    D3DXMatrixMultiply(m, m_matWorld, m_matView);
    D3DXMatrixMultiplyTranspose(m, m, m_matProj);
    m_pd3dDevice.SetVertexShaderConstantF(0, PSingle(@m), 4);

    // Set the scale factor for the displacement mapping
    // float c[4] = {0.15f,0,0,0};
    m_pd3dDevice.SetVertexShaderConstantF(4, @c, 1);

    // Displacement mapping only works with N-Patches
    m_pd3dDevice.SetNPatchMode(4);

    m_pd3dDevice.SetVertexDeclaration(m_pDispMapVertexDecl);
    if (m_bBumpMapOn and m_bEnvMapOn) then
    begin
      m_pd3dDevice.SetVertexShader(m_pDispMapBumpShader);
    end else
    begin
      m_pd3dDevice.SetVertexShader(m_pDispMapShader);
    end;
  end else
  begin
    m_pd3dDevice.SetTextureStageState(0, D3DTSS_TEXCOORDINDEX, 1);
    m_pd3dDevice.SetFVF(TBumpVertex_FVF);
  end;
  m_pd3dDevice.SetTextureStageState(0, D3DTSS_COLOROP, D3DTOP_SELECTARG1);
  m_pd3dDevice.SetTextureStageState(0, D3DTSS_COLORARG1, D3DTA_TEXTURE);
  m_pd3dDevice.SetTextureStageState(0, D3DTSS_COLORARG2, D3DTA_DIFFUSE);

  if (m_bBumpMapOn and m_bEnvMapOn) then
  begin
    m_pd3dDevice.SetTexture(1, m_psBumpMap);
    m_pd3dDevice.SetTextureStageState(1, D3DTSS_TEXCOORDINDEX, 1);
    if (m_BumpMapFormat = D3DFMT_V8U8)
      then m_pd3dDevice.SetTextureStageState(1, D3DTSS_COLOROP, D3DTOP_BUMPENVMAP)
      else m_pd3dDevice.SetTextureStageState(1, D3DTSS_COLOROP, D3DTOP_BUMPENVMAPLUMINANCE);
    m_pd3dDevice.SetTextureStageState(1, D3DTSS_COLORARG1, D3DTA_TEXTURE);
    m_pd3dDevice.SetTextureStageState(1, D3DTSS_COLORARG2, D3DTA_CURRENT);
    m_pd3dDevice.SetTextureStageState(1, D3DTSS_BUMPENVMAT00, F2DW(0.5));
    m_pd3dDevice.SetTextureStageState(1, D3DTSS_BUMPENVMAT01, F2DW(0.0));
    m_pd3dDevice.SetTextureStageState(1, D3DTSS_BUMPENVMAT10, F2DW(0.0));
    m_pd3dDevice.SetTextureStageState(1, D3DTSS_BUMPENVMAT11, F2DW(0.5));
    m_pd3dDevice.SetTextureStageState(1, D3DTSS_BUMPENVLSCALE,  F2DW(4.0));
    m_pd3dDevice.SetTextureStageState(1, D3DTSS_BUMPENVLOFFSET, F2DW(0.0));

    if (m_bEnvMapOn) then
    begin
      m_pd3dDevice.SetTexture(2, m_pEnvMapTexture);
      if (not m_bDispMapOn) then
        m_pd3dDevice.SetTextureStageState(2, D3DTSS_TEXCOORDINDEX, 0);
      m_pd3dDevice.SetTextureStageState(2, D3DTSS_COLOROP, D3DTOP_ADD);
      m_pd3dDevice.SetTextureStageState(2, D3DTSS_COLORARG1, D3DTA_TEXTURE);
      m_pd3dDevice.SetTextureStageState(2, D3DTSS_COLORARG2, D3DTA_CURRENT);
    end else
      m_pd3dDevice.SetTextureStageState(2, D3DTSS_COLOROP, D3DTOP_DISABLE);
  end else
  begin
    if (m_bEnvMapOn) then
    begin
      m_pd3dDevice.SetTexture(1, m_pEnvMapTexture);
      if (not m_bDispMapOn) then
        m_pd3dDevice.SetTextureStageState(1, D3DTSS_TEXCOORDINDEX, 0);
      m_pd3dDevice.SetTextureStageState(1, D3DTSS_COLOROP, D3DTOP_ADD);
      m_pd3dDevice.SetTextureStageState(1, D3DTSS_COLORARG1, D3DTA_TEXTURE);
      m_pd3dDevice.SetTextureStageState(1, D3DTSS_COLORARG2, D3DTA_CURRENT);
    end else
      m_pd3dDevice.SetTextureStageState(1, D3DTSS_COLOROP, D3DTOP_DISABLE);

    m_pd3dDevice.SetTextureStageState(2, D3DTSS_COLOROP, D3DTOP_DISABLE );
  end;

  m_pd3dDevice.SetStreamSource(0, m_pEarthVB, 0, SizeOf(TBumpVertex));

  // Verify that the texture operations are possible on the device
  // NOTE: this should be done just once at device init time (outside of the rendering loop)
  Result := m_pd3dDevice.ValidateDevice(dwNumPasses);
  if (D3D_OK = Result) or (D3DERR_DEVICELOST = Result) then
  begin
    // ValidateDevice either succeeded, or it failed because the device
    // is lost.  In either case do nothing here, as either device is
    // able to do these operations, or the device is lost and this
    // sample will do proper device lost hanlding at the next Present().
    m_bDeviceValidationFailed := False;
  end else
  begin
    // ValidateDevice failed due to device capabilities.
    // The right thing to do when device validation fails is to try
    // a different rendering technique.  This sample just warns the user.
    m_bDeviceValidationFailed := True;
  end;

  // Finally, draw the Earth
  m_pd3dDevice.DrawPrimitive( D3DPT_TRIANGLESTRIP, 0, m_dwNumSphereVertices-2);

  if m_bDispMapOn then
  begin
    // Restore state
    m_pd3dDevice.SetNPatchMode(0);
    m_pd3dDevice.SetVertexShader(nil);
  end;

  // Restore texture stage states
  m_pd3dDevice.SetTexture(0, nil);
  m_pd3dDevice.SetTextureStageState(0, D3DTSS_TEXCOORDINDEX, 0);
  m_pd3dDevice.SetTextureStageState(0, D3DTSS_COLOROP, D3DTOP_SELECTARG1);
  m_pd3dDevice.SetTextureStageState(0, D3DTSS_COLORARG1, D3DTA_TEXTURE);
  m_pd3dDevice.SetTextureStageState(0, D3DTSS_COLORARG2, D3DTA_DIFFUSE);
  m_pd3dDevice.SetTextureStageState(0, D3DTSS_ALPHAOP, D3DTOP_DISABLE);
  m_pd3dDevice.SetTextureStageState(1, D3DTSS_COLOROP, D3DTOP_DISABLE);
  m_pd3dDevice.SetTextureStageState(1, D3DTSS_ALPHAOP, D3DTOP_DISABLE);
  m_pd3dDevice.SetTextureStageState(2, D3DTSS_COLOROP, D3DTOP_DISABLE);
  m_pd3dDevice.SetTextureStageState(2, D3DTSS_ALPHAOP, D3DTOP_DISABLE);

  // Output statistics
  m_pFont.DrawText(2,  0, D3DCOLOR_ARGB(255,255,255,0), m_strFrameStats);
  m_pFont.DrawText(2, 20, D3DCOLOR_ARGB(255,255,255,0), m_strDeviceStats);

  if (m_bDeviceValidationFailed) then
  begin
    m_pFont.DrawText(2, 40, D3DCOLOR_ARGB(255,255,0,0),
      'Warning: Device validation failed.  Rendering may not look right.');
  end;

  m_pd3dDevice.EndScene;

  Result:= S_OK;
end;

//-----------------------------------------------------------------------------
// Name: InitBumpMap()
// Desc: Converts data from m_pEarthBumpTexture into the type of bump map requested
//       as m_BumpMapFormat into m_psBumpMap.
//-----------------------------------------------------------------------------
function CMyD3DApplication.InitBumpMap: Hresult;
var
  psBumpSrc: IDirect3DTexture9;
  d3dsd: TD3DSurfaceDesc;
  d3dlr: TD3DLockedRect;

  dwSrcPitch: DWORD;
  pSrcTopRow, pSrcCurRow, pSrcBotRow: PByte;
  dwDstPitch: DWORD;
  pDstTopRow, pDstCurRow{, pDstBotRow}: PByte;

  pSrcB0: PByte;  // addr of current pixel
  pSrcB1: PByte;  // addr of pixel below current pixel, wrapping to top if necessary
  pSrcB2: PByte;  // addr of pixel above current pixel, wrapping to bottom if necessary
  pDstT : PByte;  // addr of dest pixel;

  v00: Integer; // Current pixel
  v01: Integer; // Pixel to the right of current pixel, wrapping to left edge if necessary
  vM1: Integer; // Pixel to the left of current pixel, wrapping to right edge if necessary
  v10: Integer; // Pixel one line below.
  v1M: Integer; // Pixel one line above.

  y, x: DWord;
  iDu, iDv: Longint;
  uL: Word;

begin
  psBumpSrc := m_pEarthBumpTexture;

  psBumpSrc.GetLevelDesc(0, d3dsd);
  // Create the bumpmap's surface and texture objects
  if FAILED(m_pd3dDevice.CreateTexture(d3dsd.Width, d3dsd.Height, 1, 0,
      m_BumpMapFormat, D3DPOOL_MANAGED, m_psBumpMap, nil)) then
  begin
    Result:= E_FAIL;
    Exit;
  end;

  // Fill the bits of the new texture surface with bits from
  // a private format.
  psBumpSrc.LockRect(0, d3dlr, nil, 0);
  dwSrcPitch := d3dlr.Pitch;
  pSrcTopRow := d3dlr.pBits;
  pSrcCurRow := pSrcTopRow;
  pSrcBotRow := PByte(DWord(pSrcTopRow) + (dwSrcPitch * (d3dsd.Height - 1)));

  m_psBumpMap.LockRect(0, d3dlr, nil, 0);
  dwDstPitch := d3dlr.Pitch;
  pDstTopRow := d3dlr.pBits;
  pDstCurRow := pDstTopRow;

  Assert(d3dsd.Height > 0);
  Assert(d3dsd.Width > 0);
  for y:= 0 to (d3dsd.Height - 1) do
  begin
    pSrcB0 := pSrcCurRow;

    if (y = d3dsd.Height - 1)
      then pSrcB1 := pSrcTopRow
      else pSrcB1 := PByte(DWord(pSrcCurRow) + dwSrcPitch);

    if (y = 0)
      then pSrcB2 := pSrcBotRow
      else pSrcB2 := PByte(DWord(pSrcCurRow) - dwSrcPitch);

    pDstT := pDstCurRow;

    for x:= 0 to (d3dsd.Width - 1) do
    begin
      v00 := PByte(DWord(pSrcB0) + 0)^;

      if (x = d3dsd.Width - 1)
        then v01 := pSrcCurRow^
        else v01 := PByte(DWord(pSrcB0) + 4)^;

      if (x = 0)
        then vM1 := PByte(DWord(pSrcCurRow) + (4 * (d3dsd.Width - 1)))^
        else vM1 := PByte(DWord(pSrcB0) - 4)^;
      v10 := PByte(DWord(pSrcB1) + 0)^;
      v1M := PByte(DWord(pSrcB2) + 0)^;

      iDu := (vM1-v01); // The delta-u bump value
      iDv := (v1M-v10); // The delta-v bump value

      // The luminance bump value (land masses are less shiny)
      if v00 > 1 then uL := 63 else uL := 127;

      case m_BumpMapFormat of
        D3DFMT_V8U8:
        begin
          pDstT^:= iDu; Inc(pDstT);
          pDstT^:= iDv; Inc(pDstT);
        end;

        D3DFMT_L6V5U5:
        begin
          PWord(pDstT)^ := (((iDu shr 3) and $1f) shl 0);
          PWord(pDstT)^ := PWord(pDstT)^ or (((iDv shr 3) and $1f) shl  5);
          PWord(pDstT)^ := PWord(pDstT)^ or ((( uL shr 2) and $3f) shl 10);
          Inc(pDstT, 2);
        end;

        D3DFMT_X8L8V8U8:
        begin
          pDstT^:= iDu; Inc(pDstT);
          pDstT^:= iDv; Inc(pDstT);
          pDstT^:= uL;  Inc(pDstT);
          pDstT^:= 0;   Inc(pDstT);
        end;
      end;

      // Move one pixel to the left (src is 32-bpp)
      Inc(pSrcB0, 4);
      Inc(pSrcB1, 4);
      Inc(pSrcB2, 4);
    end;

    // Move to the next line
    Inc(pSrcCurRow, dwSrcPitch);
    Inc(pDstCurRow, dwDstPitch);
  end;


  m_psBumpMap.UnlockRect(0);
  psBumpSrc.UnlockRect(0);

  Result:= S_OK;
end;

//-----------------------------------------------------------------------------
// Name: InitDeviceObjects()
// Desc: Initialize scene objects.
//-----------------------------------------------------------------------------
var
  declVS: array[0..5] of TD3DVertexElement9 =
  (
    (Stream: 0; Offset:  0; _Type: D3DDECLTYPE_FLOAT3; Method: D3DDECLMETHOD_DEFAULT; Usage: D3DDECLUSAGE_POSITION; UsageIndex: 0),
    (Stream: 0; Offset: 12; _Type: D3DDECLTYPE_FLOAT3; Method: D3DDECLMETHOD_DEFAULT; Usage: D3DDECLUSAGE_NORMAL; UsageIndex: 0),
    (Stream: 0; Offset: 24; _Type: D3DDECLTYPE_FLOAT2; Method: D3DDECLMETHOD_DEFAULT; Usage: D3DDECLUSAGE_TEXCOORD; UsageIndex: 0),
    (Stream: 0; Offset: 32; _Type: D3DDECLTYPE_FLOAT2; Method: D3DDECLMETHOD_DEFAULT; Usage: D3DDECLUSAGE_TEXCOORD; UsageIndex: 1),
    (Stream: 0; Offset: 32; _Type: D3DDECLTYPE_FLOAT2; Method: D3DDECLMETHOD_LOOKUP;  Usage: D3DDECLUSAGE_SAMPLE; UsageIndex: 0),
    ({D3DDECL_END})
  );
function CMyD3DApplication.InitDeviceObjects: HResult;
var
  strPath: array[0..MAX_PATH-1] of Char;
  adapterFormat: TD3DFormat;
  bCanDoV8U8, bCanDoL6V5U5, bCanDoL8V8U8: Boolean;
  ahMenu: HMENU;
  dwShaderFlags: DWORD;
  pEffect: ID3DXEffect;
  h: TD3DXHandle;
const
  MenuStates: array[False..True] of DWord =
    (MF_GRAYED, MF_ENABLED);
begin
  m_pFont.InitDeviceObjects(m_pd3dDevice);

  Result:= D3DAPPERR_MEDIANOTFOUND;

  if FAILED(m_pSkyBox.Create_(m_pd3dDevice, PChar('lobby_skybox.x'))) then Exit;

  // Get the path to the texture
  if FAILED(DXUtil_FindMediaFileCb(strPath, SizeOf(strPath), 'Block.bmp')) then Exit;
  // Create the texture using D3DX
  if FAILED(D3DXCreateTextureFromFileEx(m_pd3dDevice, strPath,
              D3DX_DEFAULT, D3DX_DEFAULT, D3DX_DEFAULT, 0, D3DFMT_R5G6B5,
              D3DPOOL_MANAGED, D3DX_FILTER_TRIANGLE or D3DX_FILTER_MIRROR,
              D3DX_FILTER_TRIANGLE or D3DX_FILTER_MIRROR, 0, nil, nil, m_pBlockTexture)) then Exit;

  // Get the path to the texture
  if FAILED(DXUtil_FindMediaFileCb(strPath, SizeOf(strPath), 'Earth.bmp')) then Exit;
  // Create the texture using D3DX
  if FAILED(D3DXCreateTextureFromFileEx( m_pd3dDevice, strPath,
              D3DX_DEFAULT, D3DX_DEFAULT, D3DX_DEFAULT, 0, D3DFMT_R5G6B5,
              D3DPOOL_MANAGED, D3DX_FILTER_TRIANGLE or D3DX_FILTER_MIRROR,
              D3DX_FILTER_TRIANGLE or D3DX_FILTER_MIRROR, 0, nil, nil, m_pEarthTexture)) then Exit;

  // Get the path to the texture
  if FAILED(DXUtil_FindMediaFileCb(strPath, SizeOf(strPath), 'Earthbump.bmp')) then Exit;
  // Create the texture using D3DX
  if FAILED(D3DXCreateTextureFromFileEx( m_pd3dDevice, strPath,
              D3DX_DEFAULT, D3DX_DEFAULT, D3DX_DEFAULT, 0, D3DFMT_A8R8G8B8,
              D3DPOOL_MANAGED, D3DX_FILTER_TRIANGLE or D3DX_FILTER_MIRROR,
              D3DX_FILTER_TRIANGLE or D3DX_FILTER_MIRROR, 0, nil, nil, m_pEarthBumpTexture)) then Exit;

  // Get the path to the texture
  if FAILED(DXUtil_FindMediaFileCb(strPath, SizeOf(strPath), 'EarthEnvMap.bmp')) then Exit;
  // Create the texture using D3DX
  if FAILED(D3DXCreateTextureFromFileEx( m_pd3dDevice, strPath,
              D3DX_DEFAULT, D3DX_DEFAULT, D3DX_DEFAULT, 0, D3DFMT_R5G6B5,
              D3DPOOL_MANAGED, D3DX_FILTER_TRIANGLE or D3DX_FILTER_MIRROR,
              D3DX_FILTER_TRIANGLE or D3DX_FILTER_MIRROR, 0, nil, nil, m_pEnvMapTexture)) then Exit;

  // Find out which bump map texture are supported by this device
  adapterFormat := m_d3dSettings.DisplayMode.Format;
  bCanDoV8U8   := SUCCEEDED(m_pD3D.CheckDeviceFormat(m_d3dCaps.AdapterOrdinal,
                              m_d3dCaps.DeviceType, adapterFormat,
                              0, D3DRTYPE_TEXTURE,
                              D3DFMT_V8U8)) and
                    (m_d3dCaps.TextureOpCaps and D3DTEXOPCAPS_BUMPENVMAP <> 0);

  bCanDoL6V5U5 := SUCCEEDED(m_pD3D.CheckDeviceFormat(m_d3dCaps.AdapterOrdinal,
                              m_d3dCaps.DeviceType, adapterFormat,
                              0, D3DRTYPE_TEXTURE,
                              D3DFMT_L6V5U5)) and
                    (m_d3dCaps.TextureOpCaps and D3DTEXOPCAPS_BUMPENVMAPLUMINANCE <> 0);

  bCanDoL8V8U8 := SUCCEEDED(m_pD3D.CheckDeviceFormat(m_d3dCaps.AdapterOrdinal,
                              m_d3dCaps.DeviceType, adapterFormat,
                              0, D3DRTYPE_TEXTURE,
                              D3DFMT_X8L8V8U8)) and
                    (m_d3dCaps.TextureOpCaps and D3DTEXOPCAPS_BUMPENVMAPLUMINANCE <> 0);

  if bCanDoV8U8 then        m_BumpMapFormat := D3DFMT_V8U8
  else if bCanDoL6V5U5 then m_BumpMapFormat := D3DFMT_L6V5U5
  else if bCanDoL8V8U8 then m_BumpMapFormat := D3DFMT_X8L8V8U8
  else begin
    Result:= E_FAIL;
    Exit;
  end;

  // Find out which bump map texture are supported by this device
  m_bCanDoDispMap := ((m_pd3dDevice.GetSoftwareVertexProcessing = False) and
                      (m_d3dCaps.DevCaps2 and D3DDEVCAPS2_DMAPNPATCH <> 0) and
                      (m_d3dCaps.VertexShaderVersion >= D3DVS_VERSION(1,1)) and
                      SUCCEEDED(m_pD3D.CheckDeviceFormat(m_d3dCaps.AdapterOrdinal,
                                                         m_d3dCaps.DeviceType,
                                                         adapterFormat,
                                                         D3DUSAGE_DMAP,
                                                         D3DRTYPE_TEXTURE,
                                                         D3DFMT_L8)));

  if (not m_bCanDoDispMap) then m_bDispMapOn := False;

  // Set menu states
  ahMenu:= GetMenu(m_hWnd);
  EnableMenuItem(ahMenu, IDM_U8V8,   MenuStates[bCanDoV8U8]);
  EnableMenuItem(ahMenu, IDM_U5V5L6, MenuStates[bCanDoL6V5U5]);
  EnableMenuItem(ahMenu, IDM_U8V8L8, MenuStates[bCanDoL8V8U8]);
  EnableMenuItem(ahMenu, IDM_DISPMAPTOGGLE, MenuStates[m_bCanDoDispMap]);
  SetMenuStates;

  // Initialize earth geometry
  if FAILED(CreateEarthVertexBuffer) then
  begin
    Result:= E_FAIL;
    Exit;
  end;

  // Create and fill the bumpmap
  if FAILED(InitBumpMap) then
  begin
    Result:= E_FAIL;
    Exit;
  end;

  if (m_bCanDoDispMap) then
  begin
    // Create the displacement map texture
    Result := DXUtil_FindMediaFileCb( strPath, SizeOf(strPath), 'Earthbump.bmp');
    if FAILED(Result) then Exit;

    Result := D3DXCreateTextureFromFileEx(m_pd3dDevice,
                                      strPath,
                                      D3DX_DEFAULT,         // width
                                      D3DX_DEFAULT,         // height
                                      D3DX_DEFAULT,         // Num mip levels
                                      D3DUSAGE_DMAP,
                                      D3DFMT_L8,
                                      D3DPOOL_MANAGED,
                                      D3DX_DEFAULT,         // Filter
                                      D3DX_DEFAULT,         // Mip filter
                                      0,                    // Color key
                                      nil,                 // pSrcInfo
                                      nil,                 // pPalette
                                      m_pDispMapTexture);
    if FAILED(Result) then
    begin
      Result:= D3DAPPERR_MEDIANOTFOUND;
      Exit;
    end;

    // Two vertex shaders are used - one for bump mapped earth and another when bump mapping
    // is off. This is done because the texture coordinate index setup is different for
    // these cases.

    // Load effect file
    Result := DXUtil_FindMediaFileCch(strPath, MAX_PATH, 'EarthDispMap.fx');
    if FAILED(Result) then Exit;

    // Define DEBUG_VS and/or DEBUG_PS to debug vertex and/or pixel shaders with the shader debugger.
    // Debugging vertex shaders requires either REF or software vertex processing, and debugging
    // pixel shaders requires REF.  The D3DXSHADER_FORCE_*_SOFTWARE_NOOPT flag improves the debug
    // experience in the shader debugger.  It enables source level debugging, prevents instruction
    // reordering, prevents dead code elimination, and forces the compiler to compile against the next
    // higher available software target, which ensures that the unoptimized shaders do not exceed
    // the shader model limitations.  Setting these flags will cause slower rendering since the shaders
    // will be unoptimized and forced into software.  See the DirectX documentation for more information
    // about using the shader debugger.
    dwShaderFlags := 0;
    {$IFDEF DEBUG_VS}
      dwShaderFlags := dwShaderFlags or D3DXSHADER_FORCE_VS_SOFTWARE_NOOPT;
    {$ENDIF}
    {$IFDEF DEBUG_PS}
      dwShaderFlags := dwShaderFlags or D3DXSHADER_FORCE_PS_SOFTWARE_NOOPT;
    {$ENDIF}

    Result := D3DXCreateEffectFromFile(m_pd3dDevice, strPath, nil, nil,
                                       dwShaderFlags, nil, pEffect, nil);
    if FAILED(Result) then Exit;

    h := pEffect.GetParameterByName(nil, 'EarthBumpMap');
    if (h = nil) then
    begin
      SAFE_RELEASE(pEffect);
      Result:= E_FAIL;
      Exit;
    end;

    Result := pEffect.GetVertexShader(h, m_pDispMapBumpShader);
    if FAILED(Result) then Exit;

    h := pEffect.GetParameterByName(nil, 'EarthNoBumpMap');
    if (h = nil) then
    begin
      SAFE_RELEASE( pEffect);
      Result:= E_FAIL;
      Exit;
    end;

    Result := pEffect.GetVertexShader(h, m_pDispMapShader);
    if FAILED(Result) then
    begin
      SAFE_RELEASE(pEffect);
      Exit;
    end;

    SAFE_RELEASE(pEffect);

    // Earth vertex declaration to use with displacement mapping

    declVS[5]:= D3DDECL_END;
    Result := m_pd3dDevice.CreateVertexDeclaration(@declVS, m_pDispMapVertexDecl);
    if FAILED(Result) then Exit;
  end;

  m_bDeviceValidationFailed := False;

  Result:= S_OK;
end;

//-----------------------------------------------------------------------------
// Name: RestoreDeviceObjects()
// Desc: Initialize scene objects.
//-----------------------------------------------------------------------------
function CMyD3DApplication.RestoreDeviceObjects: HResult;
var
  fAspect: Single;
  vEyePt, vLookatPt: TD3DXVector3;
begin
  m_pFont.RestoreDeviceObjects;
  m_pSkyBox.RestoreDeviceObjects(m_pd3dDevice);

  m_pd3dDevice.SetSamplerState(0, D3DSAMP_ADDRESSU, D3DTADDRESS_WRAP);
  m_pd3dDevice.SetSamplerState(0, D3DSAMP_ADDRESSV, D3DTADDRESS_WRAP);
  m_pd3dDevice.SetSamplerState(0, D3DSAMP_MAGFILTER, D3DTEXF_LINEAR);
  m_pd3dDevice.SetSamplerState(0, D3DSAMP_MINFILTER, D3DTEXF_LINEAR);
  m_pd3dDevice.SetSamplerState(1, D3DSAMP_ADDRESSU, D3DTADDRESS_WRAP);
  m_pd3dDevice.SetSamplerState(1, D3DSAMP_ADDRESSV, D3DTADDRESS_WRAP);
  m_pd3dDevice.SetSamplerState(1, D3DSAMP_MAGFILTER, D3DTEXF_LINEAR);
  m_pd3dDevice.SetSamplerState(1, D3DSAMP_MINFILTER, D3DTEXF_LINEAR);
  m_pd3dDevice.SetSamplerState(2, D3DSAMP_ADDRESSU, D3DTADDRESS_WRAP);
  m_pd3dDevice.SetSamplerState(2, D3DSAMP_ADDRESSV, D3DTADDRESS_WRAP);
  m_pd3dDevice.SetSamplerState(2, D3DSAMP_MAGFILTER, D3DTEXF_LINEAR);
  m_pd3dDevice.SetSamplerState(2, D3DSAMP_MINFILTER, D3DTEXF_LINEAR);

  m_pd3dDevice.SetRenderState(D3DRS_ZENABLE, iTrue);

  // Get the aspect ratio
  fAspect := m_d3dsdBackBuffer.Width / m_d3dsdBackBuffer.Height;

  // Set projection matrix
  m_Camera.SetProjParams(D3DX_PI/4, fAspect, 0.1, 225.0);
  m_Camera.SetWindow(m_d3dsdBackBuffer.Width, m_d3dsdBackBuffer.Height);
  m_matProj := m_Camera.GetProjMatrix^;
  m_pd3dDevice.SetTransform(D3DTS_PROJECTION, m_matProj);

  // Set the ArcBall parameters
  vEyePt    := D3DXVector3(0.0, 0.0, -3.0);
  vLookatPt := D3DXVector3Zero; //(0.0, 0.0, 0.0);
  m_Camera.SetViewParams(vEyePt, vLookatPt);
  m_Camera.SetRadius(5.0, 2.0, 10.0);

  Result:= S_OK;
end;

//-----------------------------------------------------------------------------
// Name: InvalidateDeviceObjects()
// Desc:
//-----------------------------------------------------------------------------
function CMyD3DApplication.InvalidateDeviceObjects: HResult;
begin
  m_pFont.InvalidateDeviceObjects;
  m_pSkyBox.InvalidateDeviceObjects;
  Result:= S_OK;
end;

//-----------------------------------------------------------------------------
// Name: DeleteDeviceObjects()
// Desc: Called when the app is exiting, or the device is being changed,
//       this function deletes any device dependent objects.
//-----------------------------------------------------------------------------
function CMyD3DApplication.DeleteDeviceObjects: HResult;
begin
  m_pBlockTexture:= nil;
  m_pEarthTexture:= nil;
  m_pEarthBumpTexture:= nil;
  m_pEnvMapTexture:= nil;
  m_pDispMapBumpShader:= nil;
  m_pDispMapShader:= nil;
  m_pDispMapVertexDecl:= nil;
  m_pDispMapTexture:= nil;

  m_pFont.DeleteDeviceObjects;
  m_pSkyBox.Destroy_;

  m_psBumpMap:= nil;
  m_pEarthVB:= nil;
  Result:= S_OK;
end;

//-----------------------------------------------------------------------------
// Name: FinalCleanup()
// Desc: Called before the app exits, this function gives the app the chance
//       to cleanup after itself.
//-----------------------------------------------------------------------------
function CMyD3DApplication.FinalCleanup: HResult;
begin
  FreeAndNil(m_pFont);
  FreeAndNil(m_pSkyBox);

  Result:= S_OK;
end;

//-----------------------------------------------------------------------------
// Name: ConfirmDevice()
// Desc: Called during device initialization, this code checks the device
//       for some minimum set of capabilities
//-----------------------------------------------------------------------------
function CMyD3DApplication.ConfirmDevice(const pCaps: TD3DCaps9; dwBehavior: DWORD;
  adapterFormat: TD3DFormat; backBufferFormat: TD3DFormat): HRESULT; 
begin
  // Debugging vertex shaders requires either REF or software vertex processing
  // and debugging pixel shaders requires REF.
  {$IFDEF DEBUG_VS}
    if (pCaps.DeviceType <> D3DDEVTYPE_REF) and
       ((dwBehavior and D3DCREATE_SOFTWARE_VERTEXPROCESSING) = 0) then
    begin
      Result:= E_FAIL;
      Exit;
    end;
  {$ENDIF}
  {$IFDEF DEBUG_PS}
    if (pCaps.DeviceType <> D3DDEVTYPE_REF) then
    begin
      Result:= E_FAIL;
      Exit;
    end;
  {$ENDIF}

  if (dwBehavior and D3DCREATE_PUREDEVICE) <> 0 then
  begin
    Result:= E_FAIL; // GetTransform doesn't work on PUREDEVICE
    Exit;
  end;

  // Device must be able to do bumpmapping
  if (pCaps.TextureOpCaps and D3DTEXOPCAPS_BUMPENVMAPLUMINANCE) <> 0 then
  begin
    // Accept devices that can create D3DFMT_X8L8V8U8 textures
    if SUCCEEDED(m_pD3D.CheckDeviceFormat(pCaps.AdapterOrdinal,
                                          pCaps.DeviceType, adapterFormat,
                                          0, D3DRTYPE_TEXTURE,
                                          D3DFMT_X8L8V8U8)) then
    begin
      Result:= S_OK;
      Exit;
    end;

    // Accept devices that can create D3DFMT_L6V5U5 textures
    if SUCCEEDED(m_pD3D.CheckDeviceFormat(pCaps.AdapterOrdinal,
                                          pCaps.DeviceType, adapterFormat,
                                          0, D3DRTYPE_TEXTURE,
                                          D3DFMT_L6V5U5)) then
    begin
      Result:= S_OK;
      Exit;
    end;
  end;

  if (pCaps.TextureOpCaps and D3DTEXOPCAPS_BUMPENVMAP) <> 0 then
  begin
    // Accept devices that can create D3DFMT_V8U8 textures
    if SUCCEEDED(m_pD3D.CheckDeviceFormat(pCaps.AdapterOrdinal,
                                          pCaps.DeviceType, adapterFormat,
                                          0, D3DRTYPE_TEXTURE,
                                          D3DFMT_V8U8)) then
    begin
      Result:= S_OK;
      Exit;
    end;
  end;

  // Else, reject the device
  Result:= E_FAIL;
end;

//-----------------------------------------------------------------------------
// Name: CreateEarthVertexBuffer()
// Desc: Sets up the vertices for a bump-mapped sphere.
//-----------------------------------------------------------------------------
function CMyD3DApplication.CreateEarthVertexBuffer: Hresult;
var
  dwNumSphereRings:    DWord;
  dwNumSphereSegments: DWord;
  usage: DWORD;
begin
  m_pEarthVB:= nil;

  // Choose a tesselation level
  if m_bHighTesselation
    then dwNumSphereRings := 15
    else dwNumSphereRings := 5;
  if m_bHighTesselation
    then dwNumSphereSegments := 30
    else dwNumSphereSegments := 10;

  m_dwNumSphereVertices := 2 * dwNumSphereRings * (dwNumSphereSegments+1);

  usage := D3DUSAGE_WRITEONLY;
  if (m_bDispMapOn) then usage := usage or D3DUSAGE_NPATCHES;
  // Create the vertex buffer
  if FAILED(m_pd3dDevice.CreateVertexBuffer(m_dwNumSphereVertices*SizeOf(TBumpVertex),
                                            usage, TBumpVertex_FVF,
                                            D3DPOOL_MANAGED, m_pEarthVB, nil)) then
  begin
    Result:= E_FAIL;
    Exit;
  end;

  Result:= S_OK;
end;

//-----------------------------------------------------------------------------
// Name: SetMenuStates()
// Desc:
//-----------------------------------------------------------------------------
procedure CMyD3DApplication.SetMenuStates;
var
  aMenu: HMENU;
const
  MenuCheck: array[False..True] of DWord =
    (MF_UNCHECKED, MF_CHECKED);
begin
  aMenu := GetMenu(m_hWnd);

  CheckMenuItem(aMenu, IDM_TEXTURETOGGLE, MenuCheck[m_bTextureOn]);
  CheckMenuItem(aMenu, IDM_BUMPMAPTOGGLE, MenuCheck[m_bBumpMapOn]);
  CheckMenuItem(aMenu, IDM_ENVMAPTOGGLE,  MenuCheck[m_bEnvMapOn]);
  CheckMenuItem(aMenu, IDM_DISPMAPTOGGLE, MenuCheck[m_bDispMapOn]);

  CheckMenuItem(aMenu, IDM_U8V8L8, MenuCheck[m_BumpMapFormat = D3DFMT_X8L8V8U8]);
  CheckMenuItem(aMenu, IDM_U5V5L6, MenuCheck[m_BumpMapFormat = D3DFMT_L6V5U5]);
  CheckMenuItem(aMenu, IDM_U8V8,   MenuCheck[m_BumpMapFormat = D3DFMT_V8U8]);

  CheckMenuItem(aMenu, IDM_LOW_TESSELATION,  MenuCheck[not m_bHighTesselation]);
  CheckMenuItem(aMenu, IDM_HIGH_TESSELATION, MenuCheck[m_bHighTesselation]);
end;

//-----------------------------------------------------------------------------
// Name: MsgProc()
// Desc: Message proc function to handle key and menu input
//-----------------------------------------------------------------------------
function CMyD3DApplication.MsgProc(hWnd: HWND; uMsg: UINT; wParam: WPARAM;
  lParam: LPARAM): LRESULT;
begin
  // Pass mouse messages to the ArcBall so it can build internal matrices
  m_Camera.HandleMessages(hWnd, uMsg, wParam, lParam);

  // Trap context menu
  if (WM_CONTEXTMENU = uMsg) then
  begin
    Result:= 0;
    Exit;
  end;

  // Handle menu commands
  if (WM_COMMAND = uMsg) then
  begin
    case LOWORD(wParam) of
      IDM_TEXTURETOGGLE: m_bTextureOn := not m_bTextureOn;

      IDM_BUMPMAPTOGGLE: m_bBumpMapOn := not m_bBumpMapOn;

      IDM_ENVMAPTOGGLE:  m_bEnvMapOn  := not m_bEnvMapOn;

      IDM_DISPMAPTOGGLE:
      begin
        m_bDispMapOn := not m_bDispMapOn;
        CreateEarthVertexBuffer;
      end;

      IDM_U8V8L8:
      begin
        m_psBumpMap:= nil;
        m_BumpMapFormat := D3DFMT_X8L8V8U8;
        InitBumpMap;
      end;

      IDM_U5V5L6:
      begin
        m_psBumpMap:= nil;
        m_BumpMapFormat := D3DFMT_L6V5U5;
        InitBumpMap;
      end;

      IDM_U8V8:
      begin
        m_psBumpMap:= nil;
        m_BumpMapFormat := D3DFMT_V8U8;
        InitBumpMap;
      end;

      IDM_LOW_TESSELATION:
      begin
        m_bHighTesselation := FALSE;
        CreateEarthVertexBuffer;
      end;

      IDM_HIGH_TESSELATION:
      begin
        m_bHighTesselation := TRUE;
        CreateEarthVertexBuffer;
      end;
    end;

    // Update the menus, in  any state changes occurred
    SetMenuStates;
  end;

  // Pass remaining messages to default handler
  Result:= inherited MsgProc(hWnd, uMsg, wParam, lParam);
end;

end.

