(*----------------------------------------------------------------------------*
 *  Direct3D sample from DirectX 9.0 SDK Summer 2003                          *
 *  Delphi adaptation by Alexey Barkovoy (e-mail: clootie@ixbt.com)           *
 *                                                                            *
 *  Delphi versions 5-7 are supported                                         *
 *                                                                            *
 *  Modified: 10-Jan-2004                                                     *
 *                                                                            *
 *  Latest version can be downloaded from:                                    *
 *     http://clootie.narod.ru/delphi                                         *
 *----------------------------------------------------------------------------*)
//-----------------------------------------------------------------------------
// File: BumpLens.cpp
//
// Desc: Code to simulate a magnifying glass using bumpmapping.
//
// Note: Based on a sample from the Matrox web site
//
// Copyright (c) Microsoft Corporation. All rights reserved.
//-----------------------------------------------------------------------------

unit BumpLensUnit;

interface

uses
  Windows, Messages, SysUtils, Math,
  Direct3D9, D3DX9, D3DApp, D3DFont, D3DUtil, DXUtil;

//-----------------------------------------------------------------------------
// Function prototypes and global (or static) variables
//-----------------------------------------------------------------------------

type
  TBumpVertex = packed record // Vertex type used for bumpmap lens effect
    p: TD3DXVector3;
    tu1, tv1: Single;
    tu2, tv2: Single;
  end;
  PBumpVertexArray = ^TBumpVertexArray;
  TBumpVertexArray = array [0..10000] of TBumpVertex;

  TBackgroundVertex = packed record // Vertex type used for rendering background
    p: TD3DXVector4;
    color: DWord;
    tu, tv: Single;
  end;
  PBackgroundVertexArray = ^TBackgroundVertexArray;
  TBackgroundVertexArray = array [0..10000] of TBackgroundVertex;

const
  TBumpVertex_FVF = (D3DFVF_XYZ or D3DFVF_TEX2);
  TBackgroundVertex_FVF = (D3DFVF_XYZRHW or D3DFVF_DIFFUSE or D3DFVF_TEX1);

type
  //-----------------------------------------------------------------------------
  // Name: class CMyD3DApplication
  // Desc: Application class. The base class (CD3DApplication) provides the
  //       generic functionality needed in all Direct3D samples. CMyD3DApplication
  //       adds functionality specific to this sample program.
  //-----------------------------------------------------------------------------
  CMyD3DApplication = class(CD3DApplication)
  private
    m_pFont:            CD3DFont;

    m_pBackgroundVB: IDirect3DVertexBuffer9;
    m_pLensVB: IDirect3DVertexBuffer9;

    m_fLensX: Single;
    m_fLensY: Single;

    m_pBumpMapTexture: IDirect3DTexture9;
    m_pBackgroundTexture: IDirect3DTexture9;

    m_bDeviceValidationFailed: Boolean;

    function CreateBumpMap(iWidth, iHeight: Integer): HResult;

  protected
    function OneTimeSceneInit: HResult; override;
    function InitDeviceObjects: HResult; override;
    function RestoreDeviceObjects: HResult; override;
    function InvalidateDeviceObjects: HResult; override;
    function DeleteDeviceObjects: HResult; override;
    function Render: HResult; overload; override;
    function FrameMove: HResult; override;
    function FinalCleanup: HResult; override;
  public
    constructor Create;

    function ConfirmDevice(const pCaps: TD3DCaps9; dwBehavior: DWORD;
      adapterFormat: TD3DFormat; backBufferFormat: TD3DFormat): HRESULT; override;
 end;

implementation

function F2DW(f: Single): DWORD;
begin
  Result:= PDWord(@f)^;
end;

{ CMyD3DApplication }

//-----------------------------------------------------------------------------
// Name: CMyD3DApplication()
// Desc: Application constructor. Sets attributes for the app.
//-----------------------------------------------------------------------------
constructor CMyD3DApplication.Create;
begin
  inherited Create;

  m_strWindowTitle     := 'BumpLens: Lens Effect Using BumpMapping';
  m_d3dEnumeration.AppUsesDepthBuffer := False;

  m_pFont              := CD3DFont.Create('Arial', 12, D3DFONT_BOLD);

  m_pBumpMapTexture    := nil;
  m_pBackgroundTexture := nil;

  m_pBackgroundVB      := nil;
  m_pLensVB            := nil;
  m_bDeviceValidationFailed := False;
end;

//-----------------------------------------------------------------------------
// Name: OneTimeSceneInit()
// Desc: Called during initial app startup, this function performs all the
//       permanent initialization.
//-----------------------------------------------------------------------------
function CMyD3DApplication.OneTimeSceneInit: HResult;
begin
  Result:= S_OK;
end;

//-----------------------------------------------------------------------------
// Name: FrameMove()
// Desc: Called once per frame, the call is the entry point for animating
//       the scene.
//-----------------------------------------------------------------------------
function CMyD3DApplication.FrameMove: HResult;
begin
  // Get a triangle wave between -1 and 1
  m_fLensX := 2 * Abs(2 * ((m_fTime/2) - Floor(m_fTime/2) ) - 1 ) - 1;

  // Get a regulated sine wave between -1 and 1
  m_fLensY := 2 * Abs(Sin(m_fTime)) - 1;

  Result := S_OK;
end;

//-----------------------------------------------------------------------------
// Name: Render()
// Desc: Called once per frame, the call is the entry point for 3d
//       rendering. This function sets up render states, clears the
//       viewport, and renders the scene.
//-----------------------------------------------------------------------------
function CMyD3DApplication.Render: HResult;
var
  mat: TD3DXMatrix;
  matWorld, matView, matProj: TD3DXMatrix;
  vEyePt: TD3DXVector3;
  z: Single;
  dwNumPasses: DWord;
begin
  // Begin the scene
  if FAILED(m_pd3dDevice.BeginScene) then
  begin
    Result:= S_OK;
    Exit;
  end;

  // Render the background
  m_pd3dDevice.SetTexture(0, m_pBackgroundTexture);
  m_pd3dDevice.SetTextureStageState(0, D3DTSS_COLORARG1, D3DTA_TEXTURE);
  m_pd3dDevice.SetTextureStageState(0, D3DTSS_COLOROP,   D3DTOP_SELECTARG1);
  m_pd3dDevice.SetTextureStageState(1, D3DTSS_COLOROP,   D3DTOP_DISABLE);

  m_pd3dDevice.SetFVF(TBackgroundVertex_FVF);
  m_pd3dDevice.SetStreamSource(0, m_pBackgroundVB, 0, SizeOf(TBackgroundVertex));
  m_pd3dDevice.DrawPrimitive(D3DPT_TRIANGLESTRIP, 0, 2);

  // Render the lens
  m_pd3dDevice.SetTexture(0, m_pBumpMapTexture);
  m_pd3dDevice.SetTexture(1, m_pBackgroundTexture);

  m_pd3dDevice.SetTextureStageState(0, D3DTSS_COLOROP,   D3DTOP_BUMPENVMAP);
  m_pd3dDevice.SetTextureStageState(0, D3DTSS_COLORARG1, D3DTA_TEXTURE);
  m_pd3dDevice.SetTextureStageState(0, D3DTSS_COLORARG2, D3DTA_CURRENT);

  m_pd3dDevice.SetTextureStageState(0, D3DTSS_BUMPENVMAT00,   F2DW(0.2));
  m_pd3dDevice.SetTextureStageState(0, D3DTSS_BUMPENVMAT01,   F2DW(0.0));
  m_pd3dDevice.SetTextureStageState(0, D3DTSS_BUMPENVMAT10,   F2DW(0.0));
  m_pd3dDevice.SetTextureStageState(0, D3DTSS_BUMPENVMAT11,   F2DW(0.2));
  m_pd3dDevice.SetTextureStageState(0, D3DTSS_BUMPENVLSCALE,  F2DW(1.0));
  m_pd3dDevice.SetTextureStageState(0, D3DTSS_BUMPENVLOFFSET, F2DW(0.0));

  m_pd3dDevice.SetTextureStageState(1, D3DTSS_COLOROP,   D3DTOP_SELECTARG1);
  m_pd3dDevice.SetTextureStageState(1, D3DTSS_COLORARG1, D3DTA_TEXTURE);
  m_pd3dDevice.SetTextureStageState(1, D3DTSS_COLORARG2, D3DTA_CURRENT);

  // Generate texture coords depending on objects camera space position
  mat._11 := 0.5; mat._12 := 0.0;
  mat._21 := 0.0; mat._22 :=-0.5;
  mat._31 := 0.0; mat._32 := 0.0;
  mat._41 := 0.5; mat._42 := 0.5;

  // Scale-by-z here
  m_pd3dDevice.GetTransform(D3DTS_VIEW,       matView);
  m_pd3dDevice.GetTransform(D3DTS_PROJECTION, matProj);
  vEyePt:= D3DXVector3(matView._41, matView._42, matView._43);
  z := D3DXVec3Length(vEyePt);
  mat._11:= mat._11 * ( matProj._11 / ( matProj._33 * z + matProj._34 ));
  mat._22:= mat._22 * ( matProj._22 / ( matProj._33 * z + matProj._34 ));

  m_pd3dDevice.SetTransform(D3DTS_TEXTURE1, mat);
  m_pd3dDevice.SetTextureStageState(1, D3DTSS_TEXTURETRANSFORMFLAGS, D3DTTFF_COUNT2);
  m_pd3dDevice.SetTextureStageState(1, D3DTSS_TEXCOORDINDEX, D3DTSS_TCI_CAMERASPACEPOSITION or 1);

  // Position the lens
  D3DXMatrixTranslation(matWorld, 0.7 * (1000.0-256.0)*m_fLensX,
                                  0.7 * (1000.0-256.0)*m_fLensY,
                                  0.0);
  m_pd3dDevice.SetTransform(D3DTS_WORLD, matWorld);

  m_pd3dDevice.SetFVF(TBumpVertex_FVF);
  m_pd3dDevice.SetStreamSource(0, m_pLensVB, 0, SizeOf(TBumpVertex));

  // Verify that the texture operations are possible on the device
  // NOTE: this should be done just once at device init time (outside of the rendering loop)
  Result := m_pd3dDevice.ValidateDevice(dwNumPasses);
  if ((D3D_OK = Result) or (D3DERR_DEVICELOST = Result)) then
  begin
    // ValidateDevice either succeeded, or it failed because the device
    // is lost.  In either case do nothing here, as either device is
    // able to do these operations, or the device is lost and this
    // sample will do proper device lost hanlding at the next Present().
    m_bDeviceValidationFailed := False;
  end else
  begin
    // ValidateDevice failed due to device capabilities.
    // The right thing to do when device validation fails is to try
    // a different rendering technique.  This sample just warns the user.
    m_bDeviceValidationFailed := True;
  end;

  // Render the lens
  m_pd3dDevice.DrawPrimitive(D3DPT_TRIANGLESTRIP, 0, 2);

  // Output statistics
  m_pFont.DrawText(2,  0, D3DCOLOR_ARGB(255,255,255,0), m_strFrameStats);
  m_pFont.DrawText(2, 20, D3DCOLOR_ARGB(255,255,255,0), m_strDeviceStats);

  if m_bDeviceValidationFailed then
  begin
    m_pFont.DrawText(2, 40, D3DCOLOR_ARGB(255,255,0,0),
      'Warning: Device validation failed.  Rendering may not look right.');
  end;

  // End the scene.
  m_pd3dDevice.EndScene;

  Result := S_OK;
end;

//-----------------------------------------------------------------------------
// Name: CreateBumpmap()
// Desc: Create a bump map texture and fill its content to BUMPDUDV format
//-----------------------------------------------------------------------------
function CMyD3DApplication.CreateBumpMap(iWidth, iHeight: Integer): HResult;
var
  d3dlr: TD3DLockedRect;
  // dwDstPitch: DWord;
  pDst: PByte;
  mid: Integer;

  x0, y0: Integer;
  x1, y1: Integer;
  fDistSq00, fDistSq01, fDistSq10: Single;
  v00, v01, v10: Single;
  iDu, iDv: Single;
begin
  // Create the bumpmap's surface and texture objects
  if FAILED(m_pd3dDevice.CreateTexture(iWidth, iHeight, 1, 0,
      D3DFMT_V8U8, D3DPOOL_MANAGED, m_pBumpMapTexture, nil)) then
  begin
    Result := E_FAIL;
    Exit;
  end;

  // Fill the bumpmap texels to simulate a lens
  m_pBumpMapTexture.LockRect(0, d3dlr, nil, 0);
  mid := iWidth div 2;

  for y0:= 0 to (iHeight - 1) do
  begin
    // CHAR* pDst := (CHAR*)d3dlr.pBits + y0*d3dlr.Pitch;
    pDst:= d3dlr.pBits; Inc(pDst, y0*d3dlr.Pitch);

    for x0:= 0 to (iWidth - 1) do
    begin
      if (x0 = iWidth-1)  then x1:= x0 else x1:= x0+1;
      if (y0 = iHeight-1) then y1:= y0 else y1:= y0+1;

      fDistSq00 := ( (x0-mid)*(x0-mid) + (y0-mid)*(y0-mid));
      fDistSq01 := ( (x1-mid)*(x1-mid) + (y0-mid)*(y0-mid));
      fDistSq10 := ( (x0-mid)*(x0-mid) + (y1-mid)*(y1-mid));

      if ( fDistSq00 > (mid*mid) ) then v00:= 0 else v00:= Sqrt( (mid*mid) - fDistSq00);
      if ( fDistSq01 > (mid*mid) ) then v01:= 0 else v01:= Sqrt( (mid*mid) - fDistSq01);
      if ( fDistSq10 > (mid*mid) ) then v10:= 0 else v10:= Sqrt( (mid*mid) - fDistSq10);

      iDu := (128/D3DX_PI)*ArcTan(v00-v01); // The delta-u bump value
      iDv := (128/D3DX_PI)*ArcTan(v00-v10); // The delta-v bump value

      pDst^:= Round(iDu); Inc(pDst);
      pDst^:= Round(iDv); Inc(pDst);
    end;
  end;

  m_pBumpMapTexture.UnlockRect(0);

  Result := S_OK;
end;

//-----------------------------------------------------------------------------
// Name: InitDeviceObjects()
// Desc: Initialize scene objects
//-----------------------------------------------------------------------------
function CMyD3DApplication.InitDeviceObjects: HResult;
var
  strPath: array[0..MAX_PATH - 1] of Char;
  vLens: PBumpVertexArray;
begin
  m_pFont.InitDeviceObjects(m_pd3dDevice);

  Result := E_FAIL;

  // Get the path to the texture
  if FAILED(DXUtil_FindMediaFileCb(strPath, SizeOf(strPath), 'Lake.bmp')) then Exit;
  // Load the texture for the background image
  if FAILED(D3DXCreateTextureFromFileEx(m_pd3dDevice, strPath,
                   D3DX_DEFAULT, D3DX_DEFAULT, D3DX_DEFAULT, 0, D3DFMT_UNKNOWN,
                   D3DPOOL_MANAGED, D3DX_FILTER_TRIANGLE or D3DX_FILTER_MIRROR,
                   D3DX_FILTER_TRIANGLE or D3DX_FILTER_MIRROR, 0, nil, nil, m_pBackgroundTexture)) then Exit;

  // Create the bump map texture
  if FAILED(CreateBumpMap(256, 256)) then Exit;

  // Create a square for rendering the background
  if FAILED(m_pd3dDevice.CreateVertexBuffer(4*SizeOf(TBackgroundVertex),
                                            D3DUSAGE_WRITEONLY, TBackgroundVertex_FVF,
                                            D3DPOOL_MANAGED, m_pBackgroundVB, nil)) then Exit;

  // Create a square for rendering the lens
  if FAILED(m_pd3dDevice.CreateVertexBuffer(4*SizeOf(TBumpVertex),
                                            D3DUSAGE_WRITEONLY, TBumpVertex_FVF,
                                            D3DPOOL_MANAGED, m_pLensVB, nil)) then Exit;

  m_pLensVB.Lock(0, 0, Pointer(vLens), 0);
  vLens[0].p := D3DXVector3(-256.0,-256.0, 0.0);
  vLens[1].p := D3DXVector3(-256.0, 256.0, 0.0);
  vLens[2].p := D3DXVector3( 256.0,-256.0, 0.0);
  vLens[3].p := D3DXVector3( 256.0, 256.0, 0.0);
  vLens[0].tu1 := 0.0; vLens[0].tv1 := 1.0;
  vLens[1].tu1 := 0.0; vLens[1].tv1 := 0.0;
  vLens[2].tu1 := 1.0; vLens[2].tv1 := 1.0;
  vLens[3].tu1 := 1.0; vLens[3].tv1 := 0.0;
  m_pLensVB.Unlock;

  m_bDeviceValidationFailed := FALSE;

  Result := S_OK;
end;

//-----------------------------------------------------------------------------
// Name: RestoreDeviceObjects()
// Desc: Initialize scene objects
//-----------------------------------------------------------------------------
function CMyD3DApplication.RestoreDeviceObjects: HResult;
var
  matWorld, matView, matProj: TD3DXMatrix;
  vEyePt, vLookatPt, vUpVec: TD3DXVector3;
  fAspect: Single;

  vBackground: PBackgroundVertexArray;

  i: Integer;
begin
  m_pFont.RestoreDeviceObjects;

  // Set the transform matrices
  vEyePt    := D3DXVector3(0.0, 0.0, -2001.0);
  vLookatPt := D3DXVector3(0.0, 0.0,     0.0);
  vUpVec    := D3DXVector3(0.0, 1.0,     0.0);

  D3DXMatrixIdentity(matWorld);
  D3DXMatrixLookAtLH(matView, vEyePt, vLookatPt, vUpVec);
  fAspect := m_d3dsdBackBuffer.Width / m_d3dsdBackBuffer.Height;
  D3DXMatrixPerspectiveFovLH(matProj, D3DX_PI/4, fAspect, 1.0, 3000.0);
  m_pd3dDevice.SetTransform(D3DTS_WORLD,      matWorld);
  m_pd3dDevice.SetTransform(D3DTS_VIEW,       matView);
  m_pd3dDevice.SetTransform(D3DTS_PROJECTION, matProj);

  // Set any appropiate state
  m_pd3dDevice.SetSamplerState(0, D3DSAMP_MINFILTER, D3DTEXF_LINEAR);
  m_pd3dDevice.SetSamplerState(0, D3DSAMP_MAGFILTER, D3DTEXF_LINEAR);
  m_pd3dDevice.SetSamplerState(1, D3DSAMP_ADDRESSU,  D3DTADDRESS_CLAMP);
  m_pd3dDevice.SetSamplerState(1, D3DSAMP_ADDRESSV,  D3DTADDRESS_CLAMP);
  m_pd3dDevice.SetSamplerState(1, D3DSAMP_MINFILTER, D3DTEXF_LINEAR);
  m_pd3dDevice.SetSamplerState(1, D3DSAMP_MAGFILTER, D3DTEXF_LINEAR);
  m_pd3dDevice.SetSamplerState(1, D3DSAMP_MIPFILTER, D3DTEXF_NONE);

  // Size the background image
  m_pBackgroundVB.Lock(0, 0, Pointer(vBackground), 0);
  for i:= 0 to 3 do
  begin
    vBackground[i].p:= D3DXVector4(0.0, 0.0, 0.9, 1.0);
    vBackground[i].color := $ffffffff;
  end;
  vBackground[0].p.y := m_d3dsdBackBuffer.Height;
  vBackground[2].p.y := m_d3dsdBackBuffer.Height;
  vBackground[2].p.x := m_d3dsdBackBuffer.Width;
  vBackground[3].p.x := m_d3dsdBackBuffer.Width;
  vBackground[0].tu := 0.0; vBackground[0].tv := 1.0;
  vBackground[1].tu := 0.0; vBackground[1].tv := 0.0;
  vBackground[2].tu := 1.0; vBackground[2].tv := 1.0;
  vBackground[3].tu := 1.0; vBackground[3].tv := 0.0;
  m_pBackgroundVB.Unlock;

  Result := S_OK;
end;

//-----------------------------------------------------------------------------
// Name: InvalidateDeviceObjects()
// Desc:
//-----------------------------------------------------------------------------
function CMyD3DApplication.InvalidateDeviceObjects: HResult;
begin
  m_pFont.InvalidateDeviceObjects;
  Result := S_OK;
end;

//-----------------------------------------------------------------------------
// Name: DeleteDeviceObjects()
// Desc: Called when the app is exiting, or the device is being changed,
//       this function deletes any device dependent objects.
//-----------------------------------------------------------------------------
function CMyD3DApplication.DeleteDeviceObjects: HResult;
begin
  m_pFont.DeleteDeviceObjects;
  m_pBackgroundTexture:= nil;
  m_pBumpMapTexture:= nil;
  m_pBackgroundVB:= nil;
  m_pLensVB:= nil;

  Result := S_OK;
end;

//-----------------------------------------------------------------------------
// Name: FinalCleanup()
// Desc: Called before the app exits, this function gives the app the chance
//       to cleanup after itself.
//-----------------------------------------------------------------------------
function CMyD3DApplication.FinalCleanup: HResult;
begin
  FreeAndNil(m_pFont);
  Result:= S_OK;
end;

//-----------------------------------------------------------------------------
// Name: ConfirmDevice()
// Desc: Called during device intialization, this code checks the device
//       for some minimum set of capabilities
//-----------------------------------------------------------------------------
function CMyD3DApplication.ConfirmDevice(const pCaps: TD3DCaps9; dwBehavior: DWORD;
  adapterFormat: TD3DFormat; backBufferFormat: TD3DFormat): HRESULT; 
begin
  Result := E_FAIL;

  // GetTransform doesn't work on PUREDEVICE
  if (dwBehavior and D3DCREATE_PUREDEVICE) <> 0 then Exit;

  // Device must be able to do bumpmapping
  if (0 = (pCaps.TextureOpCaps and D3DTEXOPCAPS_BUMPENVMAP)) then Exit;

  // Accept devices that can create D3DFMT_V8U8 textures
  if SUCCEEDED(m_pD3D.CheckDeviceFormat(pCaps.AdapterOrdinal, pCaps.DeviceType,
       adapterFormat, 0, D3DRTYPE_TEXTURE, D3DFMT_V8U8)) then
  begin
    Result := S_OK;
    Exit;
  end;
end;

end.

