(*----------------------------------------------------------------------------*
 *  Direct3D sample from DirectX 9.0 SDK Summer 2003                          *
 *  Delphi adaptation by Alexey Barkovoy (e-mail: clootie@ixbt.com)           *
 *                                                                            *
 *  Delphi versions 5-7 are supported                                         *
 *                                                                            *
 *  Modified: 10-Jan-2004                                                     *
 *                                                                            *
 *  Latest version can be downloaded from:                                    *
 *     http://clootie.narod.ru/delphi                                         *
 *----------------------------------------------------------------------------*)
//-----------------------------------------------------------------------------
// File: Underwater.cpp
//
// Desc: Code to simulate underwater distortion.
//       Games could easily make use of this technique to achieve an underwater
//       effect without affecting geometry by rendering the scene to a texture
//       and applying the bump effect on a rectangle mapped with it.
//
// Note: From the Matrox web site demos
//
// Copyright (c) Microsoft Corporation. All rights reserved.
//-----------------------------------------------------------------------------

unit BumpUnderwaterUnit;

interface

uses
  Windows, Messages, SysUtils, Math, CommDlg,
  Direct3D9, D3DX9, D3DApp, D3DFont, D3DUtil, DXUtil;

//-----------------------------------------------------------------------------
// Function prototypes and global (or static) variables
//-----------------------------------------------------------------------------
function rnd: Single; { return (((FLOAT)rand()-(FLOAT)rand())/(2L*RAND_MAX)); }
function RNDbig: Single; { return (((FLOAT)rand())/RAND_MAX); }
function F2DW(f: Single): DWORD; { return *((DWORD*)&f); }

type
  TBumpVertex = packed record
    p: TD3DXVector3;
    n: TD3DXVector3;
    tu1, tv1: Single;
    tu2, tv2: Single;
  end;

  PBumpVertexArray = ^TBumpVertexArray;
  TBumpVertexArray = array [0..10000] of TBumpVertex;

const
  TBumpVertex_FVF = (D3DFVF_XYZ or D3DFVF_NORMAL or D3DFVF_TEX2);

type
  //-----------------------------------------------------------------------------
  // Name: class CMyD3DApplication
  // Desc: Application class. The base class (CD3DApplication) provides the
  //       generic functionality needed in all Direct3D samples. CMyD3DApplication
  //       adds functionality specific to this sample program.
  //-----------------------------------------------------------------------------
  CMyD3DApplication = class(CD3DApplication)
  private
    m_pFont:            CD3DFont;
    m_pWaterVB:         IDirect3DVertexBuffer9;
    m_pBumpMap:         IDirect3DTexture9;

    m_pBackgroundTexture:       IDirect3DTexture9;
    m_bDeviceValidationFailed:  Boolean;

    function CreateBumpMap: Hresult;

  protected
    function OneTimeSceneInit: HResult; override;
    function InitDeviceObjects: HResult; override;
    function RestoreDeviceObjects: HResult; override;
    function InvalidateDeviceObjects: HResult; override;
    function DeleteDeviceObjects: HResult; override;
    function Render: HResult; overload; override;
    function FrameMove: HResult; override;
    function FinalCleanup: HResult; override;
  public
    constructor Create;
    
    function ConfirmDevice(const pCaps: TD3DCaps9; dwBehavior: DWORD;
      adapterFormat: TD3DFormat; backBufferFormat: TD3DFormat): HRESULT; override;
 end;

implementation

function rnd: Single; { return (((FLOAT)rand()-(FLOAT)rand())/(2L*RAND_MAX)); }
begin
  Result:= Random - Random;
end;

function RNDbig: Single; { return (((FLOAT)rand())/RAND_MAX); }
begin
  Result:= Random;
end;

function F2DW(f: Single): DWORD;
begin
  Result:= PDWord(@f)^;
end;

{ CMyD3DApplication }

//-----------------------------------------------------------------------------
// Name: CMyD3DApplication()
// Desc: Application constructor. Sets attributes for the app.
//-----------------------------------------------------------------------------
constructor CMyD3DApplication.Create;
begin
  inherited Create;

  m_strWindowTitle     := 'BumpUnderWater: Effect Using BumpMapping';
  m_d3dEnumeration.AppUsesDepthBuffer := True;

  m_pFont              := CD3DFont.Create('Arial', 12, D3DFONT_BOLD);
  m_pWaterVB           := nil;
  m_pBumpMap           := nil;
  m_pBackgroundTexture := nil;
  m_bDeviceValidationFailed := False;
end;

//-----------------------------------------------------------------------------
// Name: OneTimeSceneInit()
// Desc: Called during initial app startup, this function performs all the
//       permanent initialization.
//-----------------------------------------------------------------------------
function CMyD3DApplication.OneTimeSceneInit: HResult;
begin
  Result:= S_OK;
end;

//-----------------------------------------------------------------------------
// Name: FrameMove()
// Desc: Called once per frame, the call is the entry point for animating
//       the scene.
//-----------------------------------------------------------------------------
function CMyD3DApplication.FrameMove: HResult;
var
  vWaterVertices: PBumpVertexArray;
begin
  m_pd3dDevice.SetTextureStageState(0, D3DTSS_BUMPENVMAT00, F2DW(0.01));
  m_pd3dDevice.SetTextureStageState(0, D3DTSS_BUMPENVMAT01, F2DW(0.00));
  m_pd3dDevice.SetTextureStageState(0, D3DTSS_BUMPENVMAT10, F2DW(0.00));
  m_pd3dDevice.SetTextureStageState(0, D3DTSS_BUMPENVMAT11, F2DW(0.01));

  m_pWaterVB.Lock( 0, 0, Pointer(vWaterVertices), 0);
  vWaterVertices[0].tu1 :=  0.000; vWaterVertices[0].tv1 := 0.5*m_fTime + 2.0;
  vWaterVertices[1].tu1 :=  0.000; vWaterVertices[1].tv1 := 0.5*m_fTime;
  vWaterVertices[2].tu1 :=  1.000; vWaterVertices[2].tv1 := 0.5*m_fTime;
  vWaterVertices[3].tu1 :=  1.000; vWaterVertices[3].tv1 := 0.5*m_fTime + 2.0;
  m_pWaterVB.Unlock;

  Result := S_OK;
end;

//-----------------------------------------------------------------------------
// Name: Render()
// Desc: Called once per frame, the call is the entry point for 3d
//       rendering. This function sets up render states, clears the
//       viewport, and renders the scene.
//-----------------------------------------------------------------------------
function CMyD3DApplication.Render: HResult;
var
  dwNumPasses: DWord;
begin
  // Clear the scene
  m_pd3dDevice.Clear(0, nil, D3DCLEAR_TARGET or D3DCLEAR_ZBUFFER, $ff000000, 1.0, 0);

  // Begin the scene
  if FAILED(m_pd3dDevice.BeginScene) then
  begin
    Result := S_OK;
    Exit;
  end;

  // Render the waves
  m_pd3dDevice.SetTexture(0, m_pBumpMap);
  m_pd3dDevice.SetTexture(1, m_pBackgroundTexture);

  m_pd3dDevice.SetTextureStageState(0, D3DTSS_TEXCOORDINDEX, 0);
  m_pd3dDevice.SetTextureStageState(0, D3DTSS_COLOROP,   D3DTOP_BUMPENVMAP);
  m_pd3dDevice.SetTextureStageState(0, D3DTSS_COLORARG1, D3DTA_TEXTURE);
  m_pd3dDevice.SetTextureStageState(0, D3DTSS_COLORARG2, D3DTA_CURRENT);
  m_pd3dDevice.SetTextureStageState(0, D3DTSS_ALPHAOP,   D3DTOP_DISABLE);

  m_pd3dDevice.SetTextureStageState(1, D3DTSS_TEXCOORDINDEX, 1);
  m_pd3dDevice.SetTextureStageState(1, D3DTSS_COLOROP,   D3DTOP_SELECTARG1);
  m_pd3dDevice.SetTextureStageState(1, D3DTSS_COLORARG1, D3DTA_TEXTURE);
  m_pd3dDevice.SetTextureStageState(1, D3DTSS_COLORARG2, D3DTA_CURRENT);
  m_pd3dDevice.SetTextureStageState(1, D3DTSS_ALPHAOP,   D3DTOP_DISABLE);

  m_pd3dDevice.SetTextureStageState(2, D3DTSS_COLOROP,   D3DTOP_DISABLE);
  m_pd3dDevice.SetTextureStageState(2, D3DTSS_ALPHAOP,   D3DTOP_DISABLE);

  m_pd3dDevice.SetFVF(TBumpVertex_FVF);
  m_pd3dDevice.SetStreamSource(0, m_pWaterVB, 0, SizeOf(TBumpVertex));

  // Verify that the texture operations are possible on the device
  // NOTE: this should be done just once at device init time (outside of the rendering loop)
  Result := m_pd3dDevice.ValidateDevice(dwNumPasses);
  if (D3D_OK = Result) or (D3DERR_DEVICELOST = Result) then
  begin
    // ValidateDevice either succeeded, or it failed because the device
    // is lost.  In either case do nothing here, as either device is
    // able to do these operations, or the device is lost and this
    // sample will do proper device lost hanlding at the next Present().
    m_bDeviceValidationFailed := False;
  end else
  begin
    // ValidateDevice failed due to device capabilities.
    // The right thing to do when device validation fails is to try
    // a different rendering technique.  This sample just warns the user.
    m_bDeviceValidationFailed:= True;
  end;

  m_pd3dDevice.DrawPrimitive(D3DPT_TRIANGLESTRIP, 0, 2);

  // Output statistics
  m_pFont.DrawText(2,  0, D3DCOLOR_ARGB(255,255,255,0), m_strFrameStats);
  m_pFont.DrawText(2, 20, D3DCOLOR_ARGB(255,255,255,0), m_strDeviceStats);

  if m_bDeviceValidationFailed then
  begin
    m_pFont.DrawText(2, 40, D3DCOLOR_ARGB(255,255,0,0),
      'Warning: Device validation failed.  Rendering may not look right.');
  end;

  // End the scene.
  m_pd3dDevice.EndScene;

  Result := S_OK;
end;

//-----------------------------------------------------------------------------
// Name: CreateBumpMapFromSurface()
// Desc: Creates a bumpmap from a surface
//-----------------------------------------------------------------------------
function CMyD3DApplication.CreateBumpMap: Hresult;
var
  iWidth, iHeight: Integer;
  lrDst: TD3DLockedRect;

  pDst: PByteArray;

  x, y: Integer;
  fx, fy: Single;
  iDu, iDv: Byte;
begin
  iWidth  := 256;
  iHeight := 256;

  // Create the bumpmap's surface and texture objects
  if FAILED(m_pd3dDevice.CreateTexture(iWidth, iHeight, 1, 0, D3DFMT_V8U8,
      D3DPOOL_MANAGED, m_pBumpMap, nil)) then
  begin
    Result:= E_FAIL;
    Exit;
  end;

  // Fill the bumpmap texels to simulate a lens
  m_pBumpMap.LockRect(0, lrDst, nil, 0);
  pDst:= PByteArray(lrDst.pBits);

  for y:=0 to (iHeight - 1) do
  begin
    for x:= 0 to (iWidth-1) do
    begin
      fx := x/iWidth  - 0.5;
      fy := y/iHeight - 0.5;

      iDu:= Round(64*Cos(4.0*(fx+fy)*D3DX_PI));
      iDv:= Round(64*Sin(4.0*(fx+fy)*D3DX_PI));

      pDst[2*x+0]:= iDu;
      pDst[2*x+1]:= iDv;
    end;
    Inc(PByte(pDst), lrDst.Pitch);
  end;

  m_pBumpMap.UnlockRect(0);

  Result := S_OK;
end;

//-----------------------------------------------------------------------------
// Name: InitDeviceObjects()
// Desc: Initialize scene objects
//-----------------------------------------------------------------------------
function CMyD3DApplication.InitDeviceObjects: HResult;
var
  v: PBumpVertexArray;
  strPath: array[0..MAX_PATH-1] of Char;
begin
  Result := E_FAIL;

  m_pFont.InitDeviceObjects(m_pd3dDevice);

  // Get the path to the texture
  if FAILED(DXUtil_FindMediaFileCb(strPath, SizeOf(strPath), 'LobbyXPos.jpg')) then Exit;
  // Create the texture using D3DX
  if FAILED(D3DXCreateTextureFromFileEx( m_pd3dDevice, strPath,
              D3DX_DEFAULT, D3DX_DEFAULT, D3DX_DEFAULT, 0, D3DFMT_A8R8G8B8,
              D3DPOOL_MANAGED, D3DX_FILTER_TRIANGLE or D3DX_FILTER_MIRROR,
              D3DX_FILTER_TRIANGLE or D3DX_FILTER_MIRROR, 0, nil, nil, m_pBackgroundTexture)) then Exit;

  // create a bumpmap from info in source surface
  if FAILED(CreateBumpMap)
  then Exit;

  if FAILED(m_pd3dDevice.CreateVertexBuffer(4*SizeOf(TBumpVertex),
      D3DUSAGE_WRITEONLY, TBumpVertex_FVF, D3DPOOL_MANAGED, m_pWaterVB, nil))
  then Exit;

  m_pWaterVB.Lock(0, 0, Pointer(v), 0);
  v[0].p:= D3DXVector3(-60.0,-60.0, 0.0);  v[0].n:= D3DXVector3(0, 1, 0);
  v[1].p:= D3DXVector3(-60.0, 60.0, 0.0);  v[1].n:= D3DXVector3(0, 1, 0);
  v[2].p:= D3DXVector3( 60.0,-60.0, 0.0);  v[2].n:= D3DXVector3(0, 1, 0);
  v[3].p:= D3DXVector3( 60.0, 60.0, 0.0);  v[3].n:= D3DXVector3(0, 1, 0);
  v[0].tu2 := 0.000; v[0].tv2 := 1.0;
  v[1].tu2 := 0.000; v[1].tv2 := 0.0;
  v[2].tu2 := 1.000; v[2].tv2 := 1.0;
  v[3].tu2 := 1.000; v[3].tv2 := 0.0;
  m_pWaterVB.Unlock;

  m_bDeviceValidationFailed:= FALSE;

  Result := S_OK;
end;

//-----------------------------------------------------------------------------
// Name: RestoreDeviceObjects()
// Desc: Initialize scene objects
//-----------------------------------------------------------------------------
function CMyD3DApplication.RestoreDeviceObjects: HResult;
var
  matWorld, matView, matProj: TD3DXMatrix;
  vEyePt, vLookatPt, vUpVec: TD3DXVector3;
begin
  m_pFont.RestoreDeviceObjects;

  // Set the transform matrices
  vEyePt    := D3DXVector3(0.0, 0.0, -100.0);
  vLookatPt := D3DXVector3(0.0, 0.0,    0.0);
  vUpVec    := D3DXVector3(0.0, 1.0,    0.0);

  D3DXMatrixIdentity(matWorld);
  D3DXMatrixLookAtLH(matView, vEyePt, vLookatPt, vUpVec);
  D3DXMatrixPerspectiveFovLH(matProj, 1.00, 1.0, 1.0, 3000.0);
  m_pd3dDevice.SetTransform(D3DTS_WORLD,      matWorld);
  m_pd3dDevice.SetTransform(D3DTS_VIEW,       matView);
  m_pd3dDevice.SetTransform(D3DTS_PROJECTION, matProj);

  // Set any appropiate state
  m_pd3dDevice.SetRenderState(D3DRS_AMBIENT,        $ffffffff);
  m_pd3dDevice.SetRenderState(D3DRS_DITHERENABLE,   iTrue);
  m_pd3dDevice.SetRenderState(D3DRS_SPECULARENABLE, iFalse);
  m_pd3dDevice.SetTextureStageState(0, D3DTSS_COLORARG1, D3DTA_TEXTURE);
  m_pd3dDevice.SetTextureStageState(0, D3DTSS_COLORARG2, D3DTA_DIFFUSE);
  m_pd3dDevice.SetTextureStageState(0, D3DTSS_COLOROP,   D3DTOP_MODULATE);
  m_pd3dDevice.SetSamplerState(0, D3DSAMP_MINFILTER, D3DTEXF_LINEAR);
  m_pd3dDevice.SetSamplerState(0, D3DSAMP_MAGFILTER, D3DTEXF_LINEAR);
  m_pd3dDevice.SetSamplerState(1, D3DSAMP_MINFILTER, D3DTEXF_LINEAR);
  m_pd3dDevice.SetSamplerState(1, D3DSAMP_MAGFILTER, D3DTEXF_LINEAR);

  Result := S_OK;
end;

//-----------------------------------------------------------------------------
// Name: InvalidateDeviceObjects()
// Desc:
//-----------------------------------------------------------------------------
function CMyD3DApplication.InvalidateDeviceObjects: HResult;
begin
  m_pFont.InvalidateDeviceObjects;
  Result := S_OK;
end;

//-----------------------------------------------------------------------------
// Name: DeleteDeviceObjects()
// Desc: Called when the app is exiting, or the device is being changed,
//       this function deletes any device dependent objects.
//-----------------------------------------------------------------------------
function CMyD3DApplication.DeleteDeviceObjects: HResult;
begin
  m_pFont.DeleteDeviceObjects;
  m_pWaterVB:= nil;
  m_pBumpMap:= nil;
  m_pBackgroundTexture:= nil;

  Result := S_OK;
end;

//-----------------------------------------------------------------------------
// Name: FinalCleanup()
// Desc: Called before the app exits, this function gives the app the chance
//       to cleanup after itself.
//-----------------------------------------------------------------------------
function CMyD3DApplication.FinalCleanup: HResult;
begin
  FreeAndNil(m_pFont);
  Result := S_OK;
end;

//-----------------------------------------------------------------------------
// Name: ConfirmDevice()
// Desc: Called during device initialization, this code checks the device
//       for some minimum set of capabilities
//-----------------------------------------------------------------------------
function CMyD3DApplication.ConfirmDevice(const pCaps: TD3DCaps9; dwBehavior: DWORD;
  adapterFormat: TD3DFormat; backBufferFormat: TD3DFormat): HRESULT;
begin
  // Device must be able to do bumpmapping
  if (0 = (pCaps.TextureOpCaps and D3DTEXOPCAPS_BUMPENVMAP)) then
  begin
    Result := E_FAIL;
    Exit;
  end;

  // Accept devices that can create D3DFMT_V8U8 textures
  if SUCCEEDED(m_pD3D.CheckDeviceFormat(pCaps.AdapterOrdinal, pCaps.DeviceType,
       adapterFormat, 0, D3DRTYPE_TEXTURE, D3DFMT_V8U8)) then
  begin
    Result := S_OK;
    Exit;
  end;

  Result := E_FAIL;
end;

end.

